/* File: "llvm-gcc-goto-bug.c" */

/* Author: Marc Feeley <feeley@iro.umontreal.ca> */

/*
 * This program exhibits a bug in the implementation of computed gotos
 * in LLVM GCC when the -O1 command line option is used.
 *
 * The program was tested on Mac OS X 10.7 :
 *
 * bash-3.2$ uname -a
 * Darwin feeleys-macbook 11.0.0 Darwin Kernel Version 11.0.0: Sat Jun 18 12:57:44 PDT 2011; root:xnu-1699.22.73~1/RELEASE_I386 i386
 * bash-3.2$ gcc -v
 * Using built-in specs.
 * Target: i686-apple-darwin11
 * Configured with: /private/var/tmp/llvmgcc42/llvmgcc42-2335.15~25/src/configure --disable-checking --enable-werror --prefix=/Developer/usr/llvm-gcc-4.2 --mandir=/share/man --enable-languages=c,objc,c++,obj-c++ --program-prefix=llvm- --program-transform-name=/^[cg][^.-]*$/s/$/-4.2/ --with-slibdir=/usr/lib --build=i686-apple-darwin11 --enable-llvm=/private/var/tmp/llvmgcc42/llvmgcc42-2335.15~25/dst-llvmCore/Developer/usr/local --program-prefix=i686-apple-darwin11- --host=x86_64-apple-darwin11 --target=i686-apple-darwin11 --with-gxx-include-dir=/usr/include/c++/4.2.1
 * Thread model: posix
 * gcc version 4.2.1 (Based on Apple Inc. build 5658) (LLVM build 2335.15.00)
 * bash-3.2$ gcc llvm-gcc-goto-bug.c ; ./a.out
 * 0 0x103304de3
 * 1 0x103304e19
 * 11
 * bash-3.2$ gcc -O1 llvm-gcc-goto-bug.c ; ./a.out
 * 0 0x1
 * 1 0x10f8ade33
 * Segmentation fault: 11
 *
 */

#include <stdio.h>

struct state {
  int n;
  void *pc;
};

void** foo(struct state* s)
{
  /*
   * When the -O1 option is used, LLVM GCC incorrectly initializes the
   * first element of the labels array to the address 0x1, instead
   * of the address of the label AAA.
   */

  static void* labels[] = { &&AAA, &&BBB, 0};

  if (s->pc == 0)
    return labels;

 AAA:
  s->n++;
  if (s->n > 10) return 0;
  goto *s->pc;

 BBB:
  s->n += 100;
  goto AAA;
}

int main(int argc, char* argv[])
{
  int i;
  struct state s;
  void **lbls;

  s.n = 1;
  s.pc = 0;

  lbls = foo(&s);

  for (i=0; i<2; i++)
    printf("%d %p\n", i, lbls[i]);

  s.pc = lbls[0]; /* cause the computed goto to jump to AAA, thus forming a loop */

  foo(&s);

  printf("%d\n", s.n); /* should print 11 */

  return 0;
}
